# coding=utf-8
from __future__ import print_function

from suanpan.arguments import Arg, String
from suanpan.dw import dw


class DWArg(Arg):
    def fixArgValue(self, value):
        return (
            value.replace(self.ARG_VALUE_DELIMITER, "_")
            if isinstance(value, str)
            else value
        )


class Table(DWArg):
    def __init__(self, key, table, partition):
        super(Table, self).__init__(key)
        self.table = String(key=table, required=True)
        self.partition = String(key=partition)

    @property
    def isSet(self):
        return True

    def addParserArguments(self, parser):
        self.table.addParserArguments(parser)
        self.partition.addParserArguments(parser)

    def load(self, args):
        self.table.load(args)
        self.partition.load(args)
        self.value = dict(table=self.table.value, partition=self.partition.value)

    def format(self, context):
        self.value = dw.readTable(self.table.value, self.partition.value)
        return self.value

    def save(self, context, result):
        data = result.value
        dw.writeTable(self.table.value, data)
        self.logSaved(self.table.value)
        return self.value



# coding=utf-8
from __future__ import print_function

import functools
import itertools
import time
import traceback

import redis

from suanpan import utils
from suanpan.log import logger


MQ_DELIVER_KEY = "_suanpan_mq_deliver"


class RedisMQ(object):
    def __init__(
        self,
        redisHost="localhost",
        redisPort=6379,
        redisRealtime=False,
        options=None,
        client=None,
        **kwargs
    ):
        self.options = options or {}
        self.options.update(host=redisHost, port=redisPort, realtime=redisRealtime)
        if client and not isinstance(client, redis.Redis):
            raise Exception("Invalid client: {}".format(client))
        self.client = client or redis.Redis(
            host=self.options["host"], port=self.options["port"], decode_responses=True
        )

    @property
    def connected(self):
        return bool(self.client.connection)

    def createQueue(self, name, group="default", consumeID="0", force=False):
        if force:
            self.deleteQueue(name)
        return self._createQueue(name, group=group, consumeID=consumeID)

    def _createQueue(self, name, group="default", consumeID="0"):
        try:
            return self.client.xgroup_create(name, group, id=consumeID, mkstream=True)
        except Exception:
            traceback.print_exc()
            raise Exception("Queue existed: {}".format(name))

    def deleteQueue(self, *names):
        return self.client.delete(*names)

    # def hasQueue(self, name, group="default"):
    #     try:
    #         queue = self.client.xinfo_stream(name)
    #         groups = self.client.xinfo_groups(name)
    #         return any(g["name"].decode() == group for g in groups)
    #     except Exception:
    #         return False

    def sendMessage(
        self, queue, data, messageID="*", maxlen=1000, trimImmediately=False
    ):
        return self.client.xadd(
            queue, data, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
        )

    def recvMessages(
        self,
        queue,
        group="default",
        consumer="unknown",
        noack=False,
        block=None,
        count=1,
        consumeID=">",
    ):
        messages = self.client.xreadgroup(
            group, consumer, {queue: consumeID}, count=count, block=block, noack=noack
        )
        return list(self._parseMessagesGenerator(messages, group))

    def subscribeQueue(
        self,
        queue,
        group="default",
        consumer="unknown",
        noack=False,
        block=None,
        count=1,
        consumeID=">",
        delay=1,
        errDelay=1,
        errCallback=logger.error,
    ):
        while True:
            try:
                messages = self.recvMessages(
                    queue,
                    group=group,
                    consumer=consumer,
                    noack=noack,
                    block=block,
                    count=count,
                    consumeID=consumeID,
                )
            except Exception as e:
                errCallback(e)
                logger.debug("Error in receiving messages. Wait {}s".format(errDelay))
                time.sleep(errDelay)
                continue

            if not messages:
                logger.debug("Received no messages. Wait {}s".format(delay))
                time.sleep(delay)
                continue

            for message in messages:
                yield message
                self.client.xack(queue, group, message["id"])

    def recvPendingMessagesInfo(
        self,
        queue,
        group="default",
        consumer="unknown",
        start="-",
        end="+",
        count=None,
        countEveryTime=None,
    ):
        countEveryTime = countEveryTime or count or 100
        messages = []
        while True:
            msgs = self.client.xpending_range(
                queue, group, start, end, countEveryTime, consumername=consumer
            )
            messages.extend(msgs)
            if not msgs or (count and len(messages) >= count):
                return messages

    def retryPendingMessages(
        self,
        queue,
        group="default",
        consumer="unknown",
        count=100,
        maxTimes=3,
        timeout=1,
        errCallback=logger.error,
        maxlen=1000,
        trimImmediately=False,
    ):
        pendingMessages = {
            msg["id"]: msg
            for msg in self.recvMessages(
                queue, group=group, consumer=consumer, count=count, consumeID="0"
            )
        }
        pendingInfos = {
            msg["message_id"]: msg
            for msg in self.recvPendingMessagesInfo(
                queue, group=group, consumer=consumer, count=count
            )
        }
        for mid in pendingMessages.keys():
            message = pendingMessages[mid]
            info = pendingInfos.get(mid, {})
            message = utils.merge(message, info)
            data = message["data"]
            deliveredTimes = int(data.pop(MQ_DELIVER_KEY, 1))
            if deliveredTimes >= maxTimes:
                logger.error(
                    "Message {} retry failed {} times. Drop!".format(
                        message["id"], deliveredTimes
                    )
                )
                errCallback(message)
                self.client.xack(queue, group, message["id"])
                continue
            timeSinceDelivered = message.get("time_since_delivered", 0)
            if timeSinceDelivered < timeout:
                logger.warning(
                    "Message {} maybe lost: {} < {}".format(
                        message["id"], timeSinceDelivered, timeout
                    )
                )
                continue
            success = self.client.xack(queue, group, message["id"])
            if success:
                data.update({MQ_DELIVER_KEY: deliveredTimes + 1})
                self.sendMessage(
                    queue, data, maxlen=maxlen, trimImmediately=trimImmediately
                )
                logger.warning("Message send back to queue: {}".format(data))

    def _parseMessagesGenerator(self, messages, group):
        for message in messages:
            queue, items = message
            for item in items:
                mid, data = item
                if not data:
                    import pdb

                    pdb.set_trace()
                yield {"id": mid, "data": data, "queue": queue, "group": group}



# coding=utf-8
from __future__ import division, print_function

import functools
import os
import shutil
import tempfile
from threading import Thread

import tqdm
from lostc import collection as lcc
from minio import Minio

from suanpan import asyncio, path
from suanpan.log import logger
from suanpan.retry import retry
from suanpan.storage.objects import Storage


class MinioStorage(Storage):
    def __init__(
        self,
        minioAccessKey,
        minioSecretKey,
        minioBucketName="suanpan",
        minioEndpoint="localhost:9000",
        minioSecure=True,
        minioDelimiter="/",
        minioTempStore=tempfile.gettempdir(),
        **kwargs
    ):
        super(MinioStorage, self).__init__(
            delimiter=minioDelimiter, tempStore=minioTempStore, **kwargs
        )

        self.accessKey = minioAccessKey
        self.secretKey = minioSecretKey
        self.bucketName = minioBucketName
        self.endpoint = minioEndpoint
        self.bucket = minioBucketName
        self.secure = minioSecure

        self.client = Minio(
            self.endpoint,
            access_key=self.accessKey,
            secret_key=self.secretKey,
            secure=self.secure,
        )

    @retry(stop_max_attempt_number=3)
    def download(self, name, path, bucket=None, ignores=None):
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        downloadFunction = (
            self.downloadFile
            if self.isFile(name, bucket=bucket)
            else self.downloadFolder
        )
        return downloadFunction(name, path, bucket=bucket, ignores=ignores)

    def downloadFolder(
        self,
        folderName,
        folderPath,
        bucket=None,
        delimiter=None,
        workers=None,
        ignores=None,
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        delimiter = delimiter or self.delimiter
        storagePath = self.storageUrl(bucket, folderName)

        if folderPath in ignores:
            logger.info(
                "Ignore downloading folder: {} -> {}".format(folderPath, storagePath)
            )
            return folderPath

        downloads = {
            file: self.localPathJoin(
                folderPath, self.storageRelativePath(file, folderName)
            )
            for _, _, files in self.walk(folderName, delimiter=delimiter, bucket=bucket)
            for file in files
        }

        logger.info("Downloading folder: {} -> {}".format(storagePath, folderPath))
        # Download from minio
        _run = functools.partial(
            self.downloadFile, bucket=bucket, ignores=ignores, quiet=True
        )
        asyncio.starmap(
            _run, downloads.items(), pbar="Downloading", thread=True, workers=workers
        )
        # Remove ignores
        self.removeIgnores(folderPath, ignores=ignores)
        # Remove rest files and folders
        files = (
            self.localPathJoin(root, file)
            for root, _, files in os.walk(folderPath)
            for file in files
        )
        restFiles = [file for file in files if file not in downloads.values()]
        asyncio.map(
            path.remove,
            restFiles,
            pbar="Removing Rest Files" if restFiles else False,
            thread=True,
        )
        path.removeEmptyFolders(folderPath)
        logger.info("Removed empty folders in: {}".format(folderPath))
        # End
        logger.info("Downloaded folder: {} -> {}".format(storagePath, folderPath))
        return folderPath

    def downloadFile(
        self, objectName, filePath, bucket=None, ignores=None, quiet=False
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        storagePath = self.storageUrl(bucket, objectName)
        fileSize = self.getStorageSize(objectName, bucket=bucket)

        if not quiet:
            logger.info("Downloading file: {} -> {}".format(storagePath, filePath))

        with tqdm.tqdm(
            total=fileSize, bar_format=self.PBAR_FORMAT, disable=quiet
        ) as pbar:
            if filePath in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return filePath

            objectMd5 = self.getStorageMd5(objectName, bucket=bucket)
            fileMd5 = self.getLocalMd5(filePath)
            if self.checkMd5(objectMd5, fileMd5, bucket=bucket):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return filePath

            path.safeMkdirsForFile(filePath)
            self.client.fget_object(bucket, objectName, filePath)

            pbar.update(fileSize)
            pbar.set_description("Downloaded")

            return filePath

        bucket = bucket or self.bucket
        self.client.fget_object(bucket, objectName, filePath)
        return filePath

    @retry(stop_max_attempt_number=3)
    def upload(self, name, path, bucket=None, ignores=None):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        uploadFunction = self.uploadFolder if os.path.isdir(path) else self.uploadFile
        return uploadFunction(name, path, bucket=bucket, ignores=ignores)

    def uploadFolder(
        self, folderName, folderPath, bucket=None, workers=None, ignores=None
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        storagePath = self.storageUrl(bucket, folderName)

        if folderName in ignores:
            logger.info(
                "Ignore uploading folder: {} -> {}".format(folderName, storagePath)
            )
            return folderPath

        filePaths = (
            os.path.join(root, file)
            for root, _, files in os.walk(folderPath)
            for file in files
        )
        uploads = {
            filePath: self.storagePathJoin(
                folderName, self.localRelativePath(filePath, folderPath)
            )
            for filePath in filePaths
        }

        if not uploads:
            logger.warning("Uploading empty folder: {}".format(folderPath))
            return folderPath

        logger.info("Uploading folder: {} -> {}".format(folderPath, storagePath))
        # Upload files to oss
        uploadItems = [
            (objectName, filePath) for filePath, objectName in uploads.items()
        ]
        _run = functools.partial(
            self.uploadFile, bucket=bucket, ignores=ignores, quiet=True
        )
        asyncio.starmap(
            _run, uploadItems, pbar="Uploading", thread=True, workers=workers
        )
        # Remove rest files
        localFiles = set(
            self.localRelativePath(filePath, folderPath) for filePath in uploads.keys()
        )
        remoteFiles = set(
            self.storageRelativePath(objectName, folderName)
            for _, _, files in self.walk(folderName, bucket=bucket)
            for objectName in files
        )
        restFiles = [
            self.storagePathJoin(folderName, file) for file in remoteFiles - localFiles
        ]
        _run = functools.partial(self.remove, bucket=bucket, quiet=True)
        asyncio.map(
            _run,
            restFiles,
            pbar="Removing Rest Files" if restFiles else False,
            thread=True,
        )
        # End
        logger.info("Uploaded folder: {} -> {}".format(folderPath, storagePath))
        return folderPath

    def uploadFile(self, objectName, filePath, bucket=None, ignores=None, quiet=False):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        storagePath = self.storageUrl(bucket, objectName)
        fileSize = os.path.getsize(filePath)

        if not quiet:
            logger.info("Uploading file: {} -> {}".format(filePath, storagePath))

        with tqdm.tqdm(
            total=fileSize, bar_format=self.PBAR_FORMAT, disable=quiet
        ) as pbar:

            if filePath in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return filePath

            objectMd5 = self.getStorageMd5(objectName, bucket=bucket)
            fileMd5 = self.getLocalMd5(filePath)
            if self.checkMd5(objectMd5, fileMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return filePath

            self.client.fput_object(
                bucket,
                objectName,
                filePath,
                progress=Progress(pbar),
                metadata={self.CONTENT_MD5: fileMd5},
            )

            pbar.set_description("Uploaded")

            return filePath

    @retry(stop_max_attempt_number=3)
    def copy(self, name, dist, bucket=None):
        bucket = bucket or self.bucket
        copyFunction = (
            self.copyFile if self.isFile(name, bucket=bucket) else self.copyFolder
        )
        return copyFunction(name, dist, bucket=bucket)

    def copyFolder(
        self, folderName, distName, bucket=None, workers=None, delimiter=None
    ):
        bucket = bucket or self.bucket
        delimiter = delimiter or self.delimiter
        folderName = self.completePath(folderName)
        distName = self.completePath(distName)
        logger.info("Copying folder: {} -> {}".format(folderName, distName))
        copyItems = [
            (file, file.replace(folderName, distName))
            for _, _, files in self.walk(folderName, delimiter=delimiter, bucket=bucket)
            for file in files
        ]
        _run = functools.partial(self.copyFile, bucket=bucket, quiet=True)
        asyncio.starmap(_run, copyItems, pbar="Copying", thread=True, workers=workers)

    def copyFile(self, objectName, distName, bucket=None, quiet=False):
        bucket = bucket or self.bucket
        fileSize = self.getStorageSize(objectName, bucket=bucket)

        if not quiet:
            logger.info(
                "Copying file: {} -> {}".format(
                    self.storageUrl(objectName, bucket=bucket),
                    self.storageUrl(distName, bucket=bucket),
                )
            )

        with tqdm.tqdm(
            total=fileSize, bar_format=self.PBAR_FORMAT, disable=quiet
        ) as pbar:
            objectMd5 = self.getStorageMd5(objectName, bucket=bucket)
            distMd5 = self.getStorageMd5(distName, bucket=bucket)
            if self.checkMd5(objectMd5, distMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return distName

            sourcePath = self.delimiter + self.storagePathJoin(bucket, objectName)
            self.client.copy_object(bucket, distName, sourcePath)
            pbar.update(fileSize)
            return distName

    @retry(stop_max_attempt_number=3)
    def remove(self, objectName, delimiter=None, bucket=None, quiet=False):
        delimiter = delimiter or self.delimiter
        bucket = bucket or self.bucket
        removeFunc = (
            self.removeFile
            if self.isFile(objectName, bucket=bucket)
            else self.removeFolder
        )
        return removeFunc(objectName, delimiter=delimiter, bucket=bucket, quiet=quiet)

    def removeFolder(
        self, folderName, delimiter=None, bucket=None, workers=None, quiet=False
    ):
        delimiter = delimiter or self.delimiter
        bucket = bucket or self.bucket
        folderName = self.completePath(folderName)
        removes = [
            objectName
            for _, _, files in self.walk(folderName, bucket=bucket, delimiter=delimiter)
            for objectName in files
        ]
        _run = functools.partial(
            self.remove, delimiter=delimiter, bucket=bucket, quiet=True
        )
        asyncio.map(
            _run,
            removes,
            pbar="Removing" if removes and not quiet else False,
            thread=True,
            workers=workers,
        )
        return folderName

    def removeFile(self, objectName, delimiter=None, bucket=None, quiet=False):
        bucket = bucket or self.bucket
        self.client.remove_object(bucket, objectName)
        if not quiet:
            storagePath = self.storageUrl(objectName, bucket=bucket)
            logger.info("Removed file: {}".format(storagePath))
        return objectName

    def walk(self, folderName, delimiter=None, bucket=None):
        bucket = bucket or self.bucket
        delimiter = delimiter or self.delimiter
        root = self.completePath(folderName, delimiter=delimiter)
        objects = self.client.list_objects_v2(bucket, prefix=root, recursive=True)
        folders, files = lcc.divide(objects, lambda obj: obj.is_dir)
        yield root, self._getObjectNames(folders), self._getObjectNames(files)

    def listAll(self, folderName, delimiter=None, bucket=None):
        bucket = bucket or self.bucket
        delimiter = delimiter or self.delimiter
        root = self.completePath(folderName, delimiter=delimiter)
        return self.client.list_objects_v2(bucket, prefix=root, recursive=False)

    def listFolders(self, folderName, delimiter=None, bucket=None):
        return (
            obj
            for obj in self.listAll(folderName, delimiter=delimiter, bucket=bucket)
            if obj.is_dir
        )

    def listFiles(self, folderName, delimiter=None, bucket=None):
        return (
            obj
            for obj in self.listAll(folderName, delimiter=delimiter, bucket=bucket)
            if not obj.is_dir
        )

    def isFolder(self, folderName, bucket=None):
        return next(self.listAll(folderName, bucket=bucket), None)

    def isFile(self, objectName, bucket=None):
        bucket = bucket or self.bucket
        try:
            self.client.get_object(bucket, objectName)
            return True
        except Exception:
            return False

    def getStorageMd5(self, name, bucket=None):
        bucket = bucket or self.bucket
        try:
            return self.client.stat_object(bucket, name).metadata.get(self.CONTENT_MD5)
        except Exception:
            return None

    def getStorageSize(self, name, bucket=None):
        bucket = bucket or self.bucket
        return self.client.stat_object(bucket, name).size

    def storageUrl(self, path, bucket=None):
        bucket = bucket or self.bucket
        return "minio:///" + self.storagePathJoin(bucket, path)

    def _getObjectNames(self, objects):
        return (
            [obj.object_name for obj in objects]
            if isinstance(objects, list) or isinstance(objects, tuple)
            else objects.key
        )


class Progress(Thread):
    def __init__(self, pbar):
        self.pbar = pbar

    def set_meta(self, total_length, object_name):
        self.totalSize = total_length
        self.objectName = object_name

    def update(self, size):
        self.pbar.update(size)
        self.pbar.set_description("Uploading")



# coding=utf-8
from __future__ import print_function

import logging
import logging.handlers
import os
import time

from suanpan import path


class Formatter(logging.Formatter):

    converter = time.gmtime

    def __init__(
        self,
        fmt="%(asctime)s :: %(levelname)-10s :: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    ):
        super(Formatter, self).__init__(fmt=fmt, datefmt=datefmt)


class Logger(logging.Logger):

    formatter = Formatter()
    streamLogLevel = logging.DEBUG
    fileLogLevel = logging.DEBUG
    logFileMaxSize = 1024 * 1024 * 1024
    backupCount = 5
    logPath = "logs"
    logFile = None

    def __init__(self, name="suanpan"):
        super(Logger, self).__init__(name=name)
        self.streamHandler = logging.StreamHandler()
        self.streamHandler.setLevel(self.streamLogLevel)
        self.streamHandler.setFormatter(self.formatter)
        self.addHandler(self.streamHandler)

        # path.safeMkdirs(self.logPath)
        # self.logFile = "{}.log".format(name) if not self.logFile else self.logFile
        # self.fileHandler = logging.handlers.RotatingFileHandler(
        #     os.path.join(self.logPath, self.logFile),
        #     maxBytes=self.logFileMaxSize,
        #     backupCount=self.backupCount,
        # )
        # self.fileHandler.setLevel(self.fileLogLevel)
        # self.fileHandler.setFormatter(self.formatter)
        # self.addHandler(self.fileHandler)


class LoggerProxy(object):
    def __init__(self, loggerOrName):
        self.setLogger(loggerOrName)

    def __getattr__(self, key):
        return getattr(self.logger, key)

    def setLogger(self, loggerOrName):
        self.logger = (
            loggerOrName if isinstance(loggerOrName, Logger) else Logger(loggerOrName)
        )


rootLogger = Logger("suanpan")
logger = LoggerProxy(rootLogger)



#!/usr/bin/env python

"""
suanpan
"""
import os

from setuptools import find_packages, setup

BASE_REQUIRES = [
    "imageio==2.4.1",
    "numpy==1.15.2",
    "opencv-python==3.4.3.18",
    "pandas==0.23.4",
    "tqdm==4.28.1",
    "retrying==1.3.3",
    "pyodps==0.8.0",
    "tabulate==0.8.3",
    "colorama==0.4.1",
    "lostc==0.1.0",
]
HIVE_REQUIRES = [
    "sasl==0.2.1",
    "thrift-sasl==0.3.0",
    "thrift==0.11.0",
    "pyhive[hive]==0.6.1",
]
GRPC_REQUIRES = ["googleapis-common-protos==1.6.0b9", "grpcio-tools==1.19.0"]
OSS2_REQUIRES = ["oss2==2.6.1"]
MINIO_REQUIRES = ["minio==4.0.11"]
REDIS_REQUIRES = ["redis==3.2.0"]
INSTALL_REQUIRES = BASE_REQUIRES
EXTRAS_REQUIRES = {
    "service": OSS2_REQUIRES + MINIO_REQUIRES + HIVE_REQUIRES + GRPC_REQUIRES,
    "docker": OSS2_REQUIRES + MINIO_REQUIRES + HIVE_REQUIRES + REDIS_REQUIRES,
    "stream": OSS2_REQUIRES + MINIO_REQUIRES + HIVE_REQUIRES + REDIS_REQUIRES,
}
README = "README.md"


def read_file(path):
    with open(path, "r") as f:
        return f.read()


fix_packages = ["__suanpan__"]
packages = find_packages()
packages.extend(fix_packages)

setup(
    name="suanpan",
    version="0.7.48",
    packages=packages,
    license="See License",
    author="majik",
    author_email="me@yamajik.com",
    description=read_file(README),
    long_description=__doc__,
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRES,
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from datetime import date, datetime
from tempfile import NamedTemporaryFile

from py4j.protocol import Py4JJavaError
from pyspark.rdd import RDD
from pyspark.serializers import BatchedSerializer, PickleSerializer

__all__ = ["OdpsOps"]

DATE_FORMAT = "%Y-%m-%d"


def _date_to_string(d):
    if type(d) is date:
        return d.strftime(DATE_FORMAT)
    else:
        return d


def _string_to_date(s):
    try:
        d = datetime.strptime(s, DATE_FORMAT).date()
        return d
    except Exception:
        return s


class OdpsOps(object):
    def __init__(self, sc, accessId, accessKey, odpsUrl, tunnelUrl):
        self._sc = sc
        try:
            helperClass = (
                sc._jvm.java.lang.Thread.currentThread()
                .getContextClassLoader()
                .loadClass("org.apache.spark.aliyun.odps.PythonOdpsAPIHelper")
            )
            helper = helperClass.newInstance()
            self._api = helper.createPythonOdpsAPI(
                sc._jsc, accessId, accessKey, odpsUrl, tunnelUrl
            )
        except Py4JJavaError as e:
            # TODO: use --jar once it also work on driver
            if "ClassNotFoundException" in str(e.java_exception):
                OdpsOps._printErrorMsg()
            raise e
        self._gateway = sc._gateway

    def readPartitionTable(
        self,
        project,
        table,
        partition,
        numPartitions,
        cols=[],
        bytesCols=[],
        batchSize=1,
    ):
        jcols = self._to_java_array(cols)
        jbytesCols = self._to_java_array(bytesCols)
        jrdd = self._api.readTable(
            project, table, partition, jcols, jbytesCols, batchSize, numPartitions
        )
        return RDD(jrdd, self._sc, PickleSerializer())

    def readNonPartitionTable(
        self, project, table, numPartitions, cols=[], bytesCols=[], batchSize=1
    ):
        jcols = self._to_java_array(cols)
        jbytesCols = self._to_java_array(bytesCols)
        jrdd = self._api.readTable(
            project, table, jcols, jbytesCols, batchSize, numPartitions
        )
        return RDD(jrdd, self._sc, PickleSerializer())

    def saveToPartitionTable(
        self,
        project,
        table,
        partition,
        rdd,
        cols=[],
        bytesCols=[],
        isCreatePt=False,
        isOverWrite=False,
    ):
        pickledRDD = rdd._pickled()
        jcols = self._to_java_array(cols)
        jbytesCols = self._to_java_array(bytesCols)
        batched = isinstance(pickledRDD._jrdd_deserializer, BatchedSerializer)
        self._api.saveToTable(
            project,
            table,
            partition,
            pickledRDD._jrdd,
            jcols,
            jbytesCols,
            batched,
            isCreatePt,
            isOverWrite,
        )

    def saveToNonPartitionTable(self, project, table, rdd, cols=[], bytesCols=[]):
        pickledRDD = rdd._pickled()
        jcols = self._to_java_array(cols)
        jbytesCols = self._to_java_array(bytesCols)
        batched = isinstance(pickledRDD._jrdd_deserializer, BatchedSerializer)
        self._api.saveToTable(
            project, table, pickledRDD._jrdd, jcols, jbytesCols, batched
        )

    def getTableSchema(self, project, table, isPartition):
        bytesInJava = self._api.getTableSchema(project, table, isPartition).iterator()
        li = list(self._collect_iterator_through_file(bytesInJava))
        res = []
        for i in range(0, len(li), 2):
            res.append((li[i], li[i + 1]))
        return res

    def truncatTable(self, project, table):
        self._api.truncateTable(project, table)

    def getColumnByName(self, project, table, name):
        bytesInJava = self._api.getColumnByName(project, table, name).iterator()
        ret = list(self._collect_iterator_through_file(bytesInJava))
        tuple = (ret[0], ret[1])
        return tuple

    def getColumnByIdx(self, project, table, idx):
        bytesInJava = self._api.getColumnByIdx(project, table, idx).iterator()
        ret = list(self._collect_iterator_through_file(bytesInJava))
        tuple = (ret[0], ret[1])
        return tuple

    def _to_java_array(self, int_arr):
        jarray = self._gateway.new_array(self._gateway.jvm.int, len(int_arr))
        for idx, val in enumerate(int_arr):
            jarray[idx] = val
        return jarray

    def _collect_iterator_through_file(self, iterator):
        # Transferring lots of data through Py4J can be slow because
        # socket.readline() is inefficient.  Instead, we'll dump the data to a
        # file and read it back.
        tempFile = NamedTemporaryFile(delete=False, dir=self._sc._temp_dir)
        tempFile.close()
        self._api.writeToFile(iterator, tempFile.name)
        # Read the data into Python and deserialize it:
        with open(tempFile.name, "rb") as tempFile:
            for item in PickleSerializer().load_stream(tempFile):
                yield item
        os.unlink(tempFile.name)

    @staticmethod
    def _printErrorMsg():
        print(
            """
________________________________________________________________________________________________
  E-MapReduce SDK's libraries not found in class path. Try one of the following.
  1. Include the 'emr-maxcompute_2.11' library and its dependencies with in the
     spark-submit command as
     $ bin/spark-submit --packages com.aliyun.emr:emr-maxcompute_2.11:%s ...
  2. Download the JAR of the artifact from Maven Central http://search.maven.org/,
     Group Id = com.aliyun.emr, Artifact Id = emr-maxcompute_2.11, Version = %s.
     Then, include the jar in the spark-submit command as
     $ bin/spark-submit --jars <emr-maxcompute_2.11-%s.jar> ...
________________________________________________________________________________________________
"""
            % ("1.4.3", "1.4.3", "1.4.3")
        )
